from itertools import product

import pandas as pd

from numpy import isclose, float64


def switch_alleles(a):
    return {"A": "T", "T": "A", "C": "G", "G": "C"}[a]


def _palindrome(m):

    return (m.E1 == m.E2.apply(switch_alleles)) & (m.O1 == m.O2.apply(switch_alleles)) & (m.E1 == m.O1) & (m.E2 == m.O2)


def _non_inferrable_palindrome(m, tolerance):

    return _palindrome(m) & ~_exposure_and_outcome_allele_frequencies_differ_from_fifty_percent(m, tolerance)


def _inferrable_palindrome(m, tolerance):

    return _palindrome(m) & _exposure_and_outcome_allele_frequencies_differ_from_fifty_percent(m, tolerance) & _inverse_allele_frequencies_similar_within_tolerance(m, tolerance)


def _inverse_allele_frequencies_similar_within_tolerance(m, tolerance):

    return pd.Series(isclose(1 - m.OAF, m.EAF, atol=tolerance), index=m.index)


def _exposure_and_outcome_allele_frequencies_differ_from_fifty_percent(m, tolerance):

   outcome_differs = (m.OAF > 0.5 + tolerance) | (m.OAF < 0.5 - tolerance)
   exposure_differs = (m.EAF > 0.5 + tolerance) | (m.EAF < 0.5 - tolerance)

   return outcome_differs & exposure_differs


def _allele_frequencies_similar_within_tolerance(m, tolerance):

    return pd.Series(isclose(m.OAF, m.EAF, atol=tolerance), index=m.index)


def non_inferrable_palindromic_snp(m, tolerance):

    x = _non_inferrable_palindrome(m, tolerance)
    return m.loc[x].rsid


def _fine_snps(m, tolerance):

    similar_af = _allele_frequencies_similar_within_tolerance(m, tolerance)
    not_palindromic = ~_non_inferrable_palindrome(m, tolerance)

    fine = (m.E1 == m.O1) & (m.E2 == m.O2) & (m.E1 != m.E2) & similar_af & not_palindromic
    return fine


def fine_snps(m, tolerance):

    fine = _fine_snps(m, tolerance)

    return m.loc[fine].rsid


def inferrable_palindromic_snp(m, tolerance):

    x = _inferrable_palindrome(m, tolerance) & ~_fine_snps(m, tolerance)

    return m.loc[x].rsid


def _to_swap(m, tolerance):

    return (m.E1 == m.O2) & (m.E2 == m.O1) & (m.E1 != m.E2) & ~_palindromic_to_swap(m, tolerance) & ~_to_flip(m, tolerance)


def to_swap(m, tolerance):

    return m.loc[_to_swap(m, tolerance)].rsid


def palindromic_to_swap(m, tolerance):

    return m.loc[_palindromic_to_swap(m, tolerance)].rsid


def _palindromic_to_swap(m, tolerance):

    return (m.E1 == m.O2) & (m.E2 == m.O1) & (m.E1.apply(switch_alleles) == m.O1) & (m.E2.apply(switch_alleles) == m.O2) & ~_exposure_and_outcome_allele_frequencies_differ_from_fifty_percent(m, tolerance)


def to_flip(m, tolerance):

    return m.loc[_to_flip(m, tolerance)].rsid


def _to_flip(m, tolerance):

    return _allele_frequencies_similar_within_tolerance(m, tolerance) & (m.E1.apply(switch_alleles) == m.O1) & (m.E2.apply(switch_alleles) == m.O2) & ~_palindromic_to_swap(m, tolerance)


def _to_swap_and_flip(m, tolerance):

    return (m.E1.apply(switch_alleles) == m.O2) & (m.E2.apply(switch_alleles) == m.O1) & ~_fine_snps(m, tolerance) & ~_palindrome(m)


def to_flip_and_swap(m, tolerance):

    return m.loc[_to_swap_and_flip(m, tolerance)].rsid

def _incompatible(m, tolerance):

    return ~_to_flip(m, tolerance) & ~_palindromic_to_swap(m, tolerance) & ~_to_swap(m, tolerance) & ~_inferrable_palindrome(m, tolerance) & ~_non_inferrable_palindrome(m, tolerance) & ~_fine_snps(m, tolerance) & ~_to_swap_and_flip(m, tolerance)


def incompatible(m, tolerance):

    return m.loc[_incompatible(m, tolerance)].rsid


def assign_table(m, tolerance):

    m = m.copy()
    d = {
        "fine": fine_snps(m, tolerance),
        "inferrable_palindromic": inferrable_palindromic_snp(m, tolerance),
        "noninferrable_palindromic": non_inferrable_palindromic_snp(m, tolerance),
        "swapped": to_swap(m, tolerance),
        "flipped": to_flip(m, tolerance),
        "flipped_and_swapped": to_flip_and_swap(m, tolerance),
        "palindromic_swapped": palindromic_to_swap(m, tolerance),
        "incompatible": incompatible(m, tolerance)
    }

    for (li, i), (lj, j) in product(d.items(), repeat=2):
        i, j = set(i), set(j)
        if not li == lj:
            assert not i.intersection(j), "{}: {}\n{} {}\n{}".format(li, str(i), lj, str(j), set(i).intersection(j))

    m.insert(len(m.columns), "SNPType", "")
    for l, v in d.items():
        m.loc[v.index, "SNPType"] = l

    return m.set_index("rsid")


def sanity_check_table(m):

    assert m.OAF.dtype == float64
    assert m.EAF.dtype == float64

    okay = (m.OAF <= 1) & (m.OAF >= 0) & (m.EAF <= 1) & (m.EAF >= 0)
    not_okay = ~okay

    assert not_okay.sum() == 0, \
        "The following rows have invalid allele frequencies:\n" + str(m[not_okay])

    valid_alleles = "A T C G".split()
    assert all(m.E1.isin(valid_alleles)) and all(m.E2.isin(valid_alleles)) and all(m.O1.isin(valid_alleles)) and all(m.O2.isin(valid_alleles))



swap = -1
flip_and_swap = -1
palindromic_to_swap = -1

to_flip = 1

def flip_swap_remove(assigned_table, tol=0.08):

    snptype = assigned_table.SNPType.str

    swap_idx = snptype.contains("swap|^inferrable_palindromic$")
    flip_idx = snptype.contains("flip|^inferrable_palindromic$")
    remove_idx = snptype.contains("incompatible|noninferrable_palindromic")

    to_swap = list(assigned_table.ix[swap_idx].index.to_series())
    to_flip = list(assigned_table.ix[flip_idx].index.to_series())
    to_remove = list(assigned_table.ix[remove_idx].index.to_series())

    return to_swap, to_flip, to_remove


def swap_dosages(dosages, to_swap, to_remove):

    if list(dosages.columns)[0].endswith("_dose"):
        to_remove = [s + "_dose" for s in to_remove]
        to_swap = [s + "_dose" for s in to_swap]

    dosages = dosages.drop(to_remove, axis=1)
    dosages.loc[:, to_swap] = 2 - dosages.loc[:, to_swap]

    return dosages


def swap_alleles(row):
    row["REF"], row["ALT"] = row["ALT"], row["REF"]
    return row


def flip_alleles(a):
    return {"A": "T", "T": "A", "C": "G", "G": "C"}[a]


def flip_swap_remove_exposure_alleles(exposure, to_flip, to_swap, to_remove):

    exposure = exposure.drop(to_remove)

    exposure.loc[to_swap] = exposure.loc[to_swap].apply(swap_alleles, axis=1)
    exposure.loc[to_flip, "REF"] = exposure.loc[to_flip, "REF"].apply(flip_alleles)
    exposure.loc[to_flip, "ALT"] = exposure.loc[to_flip, "ALT"].apply(flip_alleles)

    return exposure



from subprocess import check_output
from io import StringIO

import pandas as pd

def vcf_to_dosage(vcf):
    "Convert vcf to dosage using bcftools"

    fixed_fields = "#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT"

    header = check_output("bcftools view -h {vcf} | tail -1".format(vcf=vcf), shell=True).decode()
    header_id_only = header.replace(fixed_fields, "ID")

    cmd = r'bcftools query -f "%ID[\t%DS]\n" {vcf}'.format(vcf=vcf)
    dosage_only = check_output(cmd, shell=True).decode()

    dosage = pd.read_table(StringIO(dosage_only), names=header_id_only.split(), sep="\t", index_col=0).T
    dosage.columns = [d + "_dose" for d in dosage.columns]
    dosage.index.name = "ID"

    return dosage



from numpy import float64

def create_exposure_outcome_table(exposure, outcome):

    exposure, outcome = exposure.copy(), outcome.copy()

    cols_to_keep_exposure = "REF ALT AF".split()
    cols_to_keep_outcome = "A1 A2 FreqA1".split()

    cols_to_drop_exposure = [c for c in exposure.columns if c not in cols_to_keep_exposure]
    cols_to_drop_outcome = [c for c in outcome.columns if c not in cols_to_keep_outcome]

    exposure = exposure.drop(cols_to_drop_exposure, 1)
    outcome = outcome.drop(cols_to_drop_outcome, 1)

    df = outcome.join(exposure)
    df.columns = "O1 O2 OAF E1 E2 EAF".split()

    df.loc[:, "O2"] = df.O2.str.upper()

    df.loc[:, ["OAF", "EAF"]] = df.loc[:, ["OAF", "EAF"]].astype(float64)

    return df["E1 E2 O1 O2 OAF EAF".split()].reset_index()



from setuptools import setup

from harmonize.version import __version__

setup(
    name = "hdose",
    packages = ["harmonize"],
    scripts = ["bin/hdose"],
    version = __version__,
    description = "Harmonize exposure, outcome and dosage for Mendelian Randomization.",
    author = "Endre Bakken Stovner",
    author_email = "endrebak85@gmail.com",
    url = "http://github.com/endrebak/harmonize",
    keywords = ["snp", "gwas", "strand", "mendelian", "randomization"],
    license = ["GPL-3.0"],
    install_requires = ["pandas>=0.16"],
    classifiers = [
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering"],
    long_description = ("Harmonize exposure, outcome and dosage for Mendelian Randomization. \n"
                        "See the URL for examples and docs.")
)
